@@ -60,14 +60,9 @@ See `built-in functionality doumentation
 details on common configuration options for sources.
 """
 
-import contextlib
 import json
 import os.path
-import shutil
 import tarfile
-import urllib.error
-import urllib.request
-from urllib.parse import urljoin
 
 # We prefer tomli that was put into standard library as tomllib
 # starting from 3.11
@@ -79,6 +74,8 @@ except ImportError:
 from buildstream import Source, SourceFetcher, SourceError
 from buildstream import utils
 
+from ._utils import download_file
+
 
 # This automatically goes into .cargo/config
 #
@@ -109,7 +106,7 @@ class Crate(SourceFetcher):
         self.name = name
         self.version = str(version)
         self.sha = sha
-        self.mark_download_url(self._get_url())
+        self.mark_download_url(cargo.url)
 
     ########################################################
     #     SourceFetcher API method implementations         #
@@ -125,9 +122,9 @@ class Crate(SourceFetcher):
             return  # pragma: nocover
 
         # Download the crate
-        crate_url = self._get_url(alias_override)
+        crate_url, auth_scheme = self._get_url(alias_override)
         with self.cargo.timed_activity("Downloading: {}".format(crate_url), silent_nested=True):
-            sha256 = self._download(crate_url)
+            sha256 = self._download(crate_url, auth_scheme)
             if self.sha is not None and sha256 != self.sha:
                 raise SourceError(
                     "File downloaded from {} has sha256sum '{}', not '{}'!".format(crate_url, sha256, self.sha)
@@ -197,67 +194,33 @@ class Crate(SourceFetcher):
     # Returns:
     #    (str): The sha256 checksum of the downloaded crate
     #
-    def _download(self, url):
+    def _download(self, url, auth_scheme):
+        # We do not use etag in case what we have in cache is
+        # not matching ref in order to be able to recover from
+        # corrupted download.
+        if self.sha:
+            etag = self._get_etag(self.sha)
+        else:
+            etag = None
 
-        try:
-            with self.cargo.tempdir() as td:
-                default_name = os.path.basename(url)
-                request = urllib.request.Request(url)
-                request.add_header("Accept", "*/*")
-                request.add_header("User-Agent", "BuildStream/2")
+        with self.cargo.tempdir() as td:
+            local_file, etag, error = download_file(url, etag, td, auth_scheme)
 
-                # We do not use etag in case what we have in cache is
-                # not matching ref in order to be able to recover from
-                # corrupted download.
-                if self.sha:
-                    etag = self._get_etag(self.sha)
-                    if etag and self.is_cached():
-                        request.add_header("If-None-Match", etag)
+            if error:
+                raise SourceError("{}: Error mirroring {}: {}".format(self, url, error), temporary=True)
 
-                with contextlib.closing(urllib.request.urlopen(request)) as response:
-                    info = response.info()
+            # Make sure url-specific mirror dir exists.
+            os.makedirs(self._get_mirror_dir(), exist_ok=True)
 
-                    etag = info["ETag"] if "ETag" in info else None
+            # Store by sha256sum
+            sha256 = utils.sha256sum(local_file)
+            # Even if the file already exists, move the new file over.
+            # In case the old file was corrupted somehow.
+            os.rename(local_file, self._get_mirror_file(sha256))
 
-                    filename = info.get_filename(default_name)
-                    filename = os.path.basename(filename)
-                    local_file = os.path.join(td, filename)
-                    with open(local_file, "wb") as dest:
-                        shutil.copyfileobj(response, dest)
-
-                # Make sure url-specific mirror dir exists.
-                os.makedirs(self._get_mirror_dir(), exist_ok=True)
-
-                # Store by sha256sum
-                sha256 = utils.sha256sum(local_file)
-                # Even if the file already exists, move the new file over.
-                # In case the old file was corrupted somehow.
-                os.rename(local_file, self._get_mirror_file(sha256))
-
-                if etag:
-                    self._store_etag(sha256, etag)
-                return sha256
-
-        except urllib.error.HTTPError as e:
-            if e.code == 304:
-                # 304 Not Modified.
-                # Because we use etag only for matching sha, currently specified sha is what
-                # we would have downloaded.
-                return self.sha
-            raise SourceError(
-                "{}: Error mirroring {}: {}".format(self, url, e),
-                temporary=True,
-            ) from e
-
-        except (
-            urllib.error.URLError,
-            urllib.error.ContentTooShortError,
-            OSError,
-        ) as e:
-            raise SourceError(
-                "{}: Error mirroring {}: {}".format(self, url, e),
-                temporary=True,
-            ) from e
+            if etag:
+                self._store_etag(sha256, etag)
+            return sha256
 
     # _get_url()
     #
@@ -270,9 +233,16 @@ class Crate(SourceFetcher):
     #    (str): The URL for this crate
     #
     def _get_url(self, alias=None):
-        url = self.cargo.translate_url(self.cargo.url, alias_override=alias)
         path = "{name}/{name}-{version}.crate".format(name=self.name, version=self.version)
-        return urljoin(f"{url}/", path)
+        extra_data = {}
+        if utils.get_bst_version() >= (2, 2):
+            translated_url = self.cargo.translate_url(
+                self.cargo.url, suffix=path, alias_override=alias, extra_data=extra_data
+            )
+        else:
+            translated_url = self.cargo.translate_url(self.cargo.url, alias_override=alias) + path
+
+        return translated_url, extra_data.get("http-auth")
 
     # _get_etag()
     #
@@ -343,10 +313,17 @@ class CargoSource(Source):
 
         # The url before any aliasing
         #
-        self.url = node.get_str("url", "https://static.crates.io/crates")
+        self.original_url = node.get_str("url", "https://static.crates.io/crates")
         self.cargo_lock = node.get_str("cargo-lock", "Cargo.lock")
         self.vendor_dir = node.get_str("vendor-dir", "crates")
 
+        # If the specified URL is just an alias, require the alias to resolve
+        # to a URL with a trailing slash. Otherwise, append a trailing slash if
+        # it's missing, for backward compatibility.
+        self.url = self.original_url
+        if not self.url.endswith(":") and not self.url.endswith("/"):
+            self.url += "/"
+
         node.validate_keys(Source.COMMON_CONFIG_KEYS + ["url", "ref", "cargo-lock", "vendor-dir"])
 
         # Needs to be marked here so that `track` can translate it later.
@@ -358,7 +335,7 @@ class CargoSource(Source):
         return
 
     def get_unique_key(self):
-        return [self.url, self.cargo_lock, self.vendor_dir, self.ref]
+        return [self.original_url, self.cargo_lock, self.vendor_dir, self.ref]
 
     def is_resolved(self):
         return (self.ref is not None) and all(crate.is_resolved() for crate in self.crates)
@@ -415,9 +392,9 @@ class CargoSource(Source):
 
             crate = Crate(self, crate_obj["name"], crate_obj["version"])
 
-            crate_url = crate._get_url()
+            crate_url, auth_scheme = crate._get_url()
             with self.timed_activity("Downloading: {}".format(crate_url), silent_nested=True):
-                crate_obj["sha"] = crate._download(crate_url)
+                crate_obj["sha"] = crate._download(crate_url, auth_scheme)
 
         return new_ref
 
