@@ -34,9 +34,13 @@ docker - stage files from Docker images
    # Specify the registry endpoint, defaults to Docker Hub (optional)
    registry-url: https://registry.hub.docker.com
 
-   # Image path (required)
+   # Image path
    image: library/alpine
 
+   # Alternatively, you can specify a full url, the registry endpoint
+   # is assumed to be at the root of the domain
+   # url: https://registry.hub.docker.com/library/alpine
+
    # Image tag to follow (optional)
    track: latest
 
@@ -121,6 +125,27 @@ def urljoin(url, *args):
     return url
 
 
+# Handles authentication with a bearer token
+class BearerAuth(requests.auth.AuthBase):
+    def __init__(self, api_timeout=3):
+        self.token = None
+        self.api_timeout = api_timeout
+
+    def __call__(self, r):
+        if self.token:
+            r.headers["Authorization"] = "Bearer {}".format(self.token)
+        return r
+
+    def refresh_token(self, auth_challenge):
+        auth_vars = parse_bearer_authorization_challenge(auth_challenge)
+        # Respond to an Www-Authenticate challenge by requesting the necessary
+        # token from the 'realm' (endpoint) that we were given in the challenge.
+        request_url = "{realm}?service={service}&scope={scope}".format(**auth_vars)
+        response = requests.get(request_url, timeout=self.api_timeout)
+        response.raise_for_status()
+        self.token = response.json()["token"]
+
+
 # DockerManifestError
 #
 # Raised if something goes wrong while querying an image manifest from a remote
@@ -137,7 +162,7 @@ class DockerRegistryV2Client:
         self.endpoint = endpoint
         self.api_timeout = api_timeout
 
-        self.token = None
+        self.auth = BearerAuth(api_timeout)
 
     def _request(self, subpath, extra_headers=None, stream=False, _reauthorized=False):
         if not extra_headers:
@@ -146,32 +171,20 @@ class DockerRegistryV2Client:
         headers = {"content-type": "application/json"}
         headers.update(extra_headers)
 
-        if self.token:
-            headers["Authorization"] = "Bearer {}".format(self.token)
-
         url = urljoin(self.endpoint, "v2", subpath)
-        response = requests.get(url, headers=headers, stream=stream, timeout=self.api_timeout)
+        response = requests.get(url, headers=headers, stream=stream, timeout=self.api_timeout, auth=self.auth)
 
         if response.status_code == requests.codes["unauthorized"] and not _reauthorized:
             # This request requires (re)authorization. See:
             # https://docs.docker.com/registry/spec/auth/token/
             auth_challenge = response.headers["Www-Authenticate"]
-            auth_vars = parse_bearer_authorization_challenge(auth_challenge)
-            self._auth(auth_vars["realm"], auth_vars["service"], auth_vars["scope"])
+            self.auth.refresh_token(auth_challenge)
             return self._request(subpath, extra_headers=extra_headers, _reauthorized=True)
         else:
             response.raise_for_status()
 
             return response
 
-    def _auth(self, realm, service, scope):
-        # Respond to an Www-Authenticate challenge by requesting the necessary
-        # token from the 'realm' (endpoint) that we were given in the challenge.
-        request_url = "{}?service={}&scope={}".format(realm, service, scope)
-        response = requests.get(request_url, timeout=self.api_timeout)
-        response.raise_for_status()
-        self.token = response.json()["token"]
-
     # digest():
     #
     # Calculate a Docker-compatible digest of an arbitrary string of bytes.
@@ -355,14 +368,24 @@ class DockerSource(Source):
             Source.COMMON_CONFIG_KEYS + ["architecture", "registry-url", "image", "os", "ref", "track", "url"]
         )
 
-        if "url" in node:
-            raise SourceError(
-                "{}: 'url' parameter is now deprecated, " "use 'registry-url' and 'image' instead.".format(self)
-            )
+        if "url" in node and ("image" in node or "registry-url" in node):
+            raise SourceError("{}: May specify either 'url', or 'image' and 'registry-url'".format(self))
 
-        self.image = node.get_str("image")
-        self.original_registry_url = node.get_str("registry-url", _DOCKER_HUB_URL)
-        self.registry_url = self.translate_url(self.original_registry_url)
+        if "url" not in node and "image" not in node:
+            raise SourceError("{}: Must define either 'url' or 'image'".format(self))
+
+        if "url" in node:
+            self.url = node.get_str("url")
+            translated_url = self.translate_url(self.url)
+            scheme, netloc, path, _, _ = urllib.parse.urlsplit(translated_url)
+            self.registry_url = urllib.parse.urlunsplit((scheme, netloc, "", "", ""))
+            self.image = path
+
+        if "image" in node:
+            self.url = None
+            self.image = node.get_str("image")
+            self.original_registry_url = node.get_str("registry-url", _DOCKER_HUB_URL)
+            self.registry_url = self.translate_url(self.original_registry_url)
 
         if "ref" in node:
             self.digest = self._ref_to_digest(node.get_str("ref"))
@@ -384,7 +407,10 @@ class DockerSource(Source):
         return
 
     def get_unique_key(self):
-        return [self.original_registry_url, self.image, self.digest]
+        if self.url is not None:
+            return [self.url, self.digest]
+        else:
+            return [self.original_registry_url, self.image, self.digest]
 
     def get_ref(self):
         return None if self.digest is None else self._digest_to_ref(self.digest)
